class YamlReaderError(Exception):
    pass


class ConfigurationError(Exception):
    pass


class RasterFileError(Exception):
    pass


class MeteoDataError(Exception):
    pass


class CategoryError(Exception):
    pass


class RuntimeError(Exception):
    pass
